"""Contains package processor class"""
import logging
from multiprocessing import Pool
import os
from pathlib import Path
import re
from typing import Dict, List, Optional, Tuple

import networkx as nx

from .exceptions import (
    CircularImportsFound,
    DuplicateTarget,
    MultipleSourcePackagesFound,
)
from .package_targets import PythonPackage
from .project import PROJECT_CONFIG
from .util import gather_dependencies_from_module, write_package_config_file

logger = logging.getLogger(__name__)

# Number of workers to use for various multiprocessing pools
PROCESSES = max(1, os.cpu_count() - 1)


class PackageProcessor:
    """Class with methods for processing internal Python packages in order to:

    * Check for circular imports
    * Generate Pants BUILD files
    * Determine package dependencies
    """

    def __init__(self) -> None:
        """Initializer"""
        # Map of target key to instance of a Python package build target
        self._targets: Dict[str, PythonPackage] = {}
        # Graph of targets
        self._target_graph: nx.DiGraph = None

    def get_target(self, target_key: str) -> PythonPackage:
        """Get a registered target by key

        Args:
            target_key: Key under which the target was registered

        Returns:
            target/package object

        """
        return self._targets[target_key]

    def register_packages(self) -> None:
        """Register targets and their dependencies.

        This should be called before performing an action with the packages.
        """
        self._register_task_targets_code()
        # self._register_task_targets_py2sfn_projects()
        # self._register_task_targets_tests()
        # self._register_extra_targets()
        self._gather_dependencies()
        self._build_target_graph()

    def _gather_dependencies(self) -> None:
        """Gather dependencies for each target"""
        logger.info("Gathering dependencies")

        # Collect a list of (build target, Python module) pairs from *all* the targets
        target_paths: List[Tuple[PythonPackage, str]] = []
        for target in self._targets.values():
            target_paths.extend(target.find_target_paths())

        # Create a pool of workers that will parse imports from each Python module
        # path.
        paths = [path for _, path in target_paths]
        with Pool(processes=PROCESSES) as pool:
            # Return a list of sets containing imported package names
            imported_package_names_per_path = pool.map(
                gather_dependencies_from_module, paths
            )

        # For each target path / package name pair, process and add the dependency to
        # the target's set of dependencies.
        for (target, _), (path, package_names) in zip(
            target_paths, imported_package_names_per_path
        ):
            for package_name in package_names:
                target.add_dependency(self._targets, path, package_name)

        # Allow each target to set extra dependencies based on their own custom logic.
        # For most targets this will be a no-op.
        for target in self._targets.values():
            target.set_extra_dependencies(self._targets)

    def generate_package_manifests(self, target_pattern: Optional[str] = None) -> None:
        """Generate package manifest files.

        You must have already called :py:meth:`.register_packages`.

        Args:
            target_pattern: If provided, pyproject.toml files are only generated for
                targets with keys matching the pattern

        """
        logger.info("Generating package manifest files")
        targets_to_save = []
        for key, target in self._targets.items():
            if not target_pattern or re.search(target_pattern, key):
                target.generate_package_manifest()
                targets_to_save.append(
                    (target.config.config_file_path, target.config.to_string())
                )

        # Create a pool of workers that will render, format, and write each manifest
        # file to disk.
        with Pool(processes=PROCESSES) as pool:
            pool.starmap(write_package_config_file, targets_to_save)

    def generate_package_manifest(self, package_path: Path) -> None:
        """Generate package manifest file for a single target.

        You must have already called :py:meth:`.register_packages`.

        Args:
            package_path: path to a Python package

        """
        p = PythonPackage(package_path)
        target = self._targets[p.package_name]
        target.generate_package_manifest()
        target.config.save_config()

    def _register_task_targets_code(self) -> None:
        """Register task targets for code packages

        This iterates through various directories looking for a setup.py file. If it
        finds one, it means we've found a Python package and can register a build
        target.
        """
        logger.info("Registering code targets")

        setup_py_paths: List[Tuple[str, Path]] = []
        for top_dir_name in PROJECT_CONFIG.top_dirs:
            top_dir = PROJECT_CONFIG.project_path.joinpath(top_dir_name)
            logger.debug(f"Walking {top_dir}")
            for dirpath, dirnames, filenames in os.walk(top_dir):
                if os.path.basename(dirpath) in PROJECT_CONFIG.ignore_dirs:
                    # Empty the list of directories so os.walk does not recur
                    dirnames.clear()
                else:
                    for filename in filenames:
                        if filename == "setup.py":
                            setup_py_paths.append(
                                (top_dir_name, Path(dirpath).joinpath(filename))
                            )

        for top_dir_name, setup_py_path in setup_py_paths:
            logger.debug(f"Registering {setup_py_path}")

            src_dir = setup_py_path.parent.joinpath("src")
            if not src_dir.is_dir():
                continue
            src_entries = [
                src_entry
                for src_entry in os.scandir(src_dir)
                if (
                    src_entry.is_dir()
                    and "egg-info" not in src_entry.path
                    and "pycache" not in src_entry.path
                )
            ]
            if len(src_entries) == 0:
                logger.debug(f"No entries found in {src_dir}")
                continue
            elif len(src_entries) > 1:
                raise MultipleSourcePackagesFound(
                    f"More than one package found in {src_dir}:"
                    f" {src_entries}. This could be caused by old git"
                    " state. Either manually clean up the packages or"
                    " run `git clean -fxd` if you are OK with a"
                    " completely clean slate."
                )

            src_entry = src_entries[0]
            package_name = os.path.basename(src_entry.path)
            target = PythonPackage(setup_py_path.parent)
            if target.key in PROJECT_CONFIG.ignore_targets:
                logger.debug(f"Ignoring {target}")
            elif package_name in self._targets:
                raise DuplicateTarget(
                    f"Duplicate target found for {package_name}. Check the"
                    " git repository for old folders that have the same"
                    " name."
                )
            else:
                self._targets[package_name] = target
                logger.debug(f"Registered target {target}")

    def _build_target_graph(self) -> None:
        """Build a graph of package targets.

        This is used to check for cycles and compute dependencies.
        """
        G = nx.DiGraph()
        for target in self._targets.values():
            G.add_node(target)
            for dependency in target.dependencies:
                G.add_edge(target, dependency)
        self._target_graph = G

    def ensure_no_circular_imports(self) -> None:
        """Ensure there are no circular imports in any of the processed packages

        You must have already called :py:meth:`.register_packages`.

        Raises:
            :py:exec:`.CircularImportsFound` if any circular imports were found

        """
        logger.info("Ensuring no circular imports")
        cycles = list(nx.simple_cycles(self._target_graph))
        if len(cycles) > 0:
            raise CircularImportsFound(cycles)



"""Contains PoetPackage class"""

from pathlib import Path
from typing import Any, Dict, MutableMapping, Optional

import toml


class PoetPackage:
    """Class that represents a package managed by ns-poet

    This object manages a Python package's pyproject.toml (i.e. manifest) file.

    The schema for the configuration file in the git project root is::

        [tool.nspoet]
        generate_package_manifest = true

    """

    def __init__(self, package_path: Path) -> None:
        """Initializer

        Args:
            package_path: Path to the package

        """
        self.package_path = package_path
        self.config_file_path = package_path.joinpath("pyproject.toml")
        self._config: Optional[MutableMapping[str, Any]] = None

    @classmethod
    def from_path(cls, package_path: Path) -> "PoetPackage":
        """Create a package object and load configuration from a given package path

        Args:
            package_path: Path to the package

        Returns:
            new PoetPackage instance

        """
        p = cls(package_path)
        p.load_config()
        return p

    def load_config(self) -> MutableMapping[str, Any]:
        """Load configuration from disk

        Returns:
            configuration dict

        """
        if self.config_file_path.is_file():
            with self.config_file_path.open() as f:
                self._config = toml.load(f)
        else:
            self._config = {}

        return self._config

    def save_config(self) -> None:
        """Save the configuration object to disk"""
        if self.generate_package_manifest:
            with self.config_file_path.open("w") as f:
                toml.dump(  # type: ignore
                    self._config,
                    f,
                    encoder=toml.encoder.TomlPreserveInlineDictEncoder(),  # type: ignore
                )

    def to_string(self) -> str:
        """Dump the configuration to a TOML string"""
        return toml.dumps(  # type: ignore
            self._config, encoder=toml.encoder.TomlPreserveInlineDictEncoder()  # type: ignore
        )

    @property
    def package_config(self) -> Dict[str, Any]:
        """Return the nspoet configuration subsection within the manifest"""
        return self._config.get("tool", {}).get("nspoet", {})

    @property
    def generate_package_manifest(self) -> bool:
        """Flag denoting whether to generate a package manifest file"""
        return self.package_config.get("generate_package_manifest", True)

    def update(
        self, name: str, dependencies: Dict[str, str], dev_dependencies: Dict[str, str]
    ) -> None:
        """Update the package configuration in place

        Args:
            name: package name
            dependencies: map of dependency name to version specifier
            dev_dependencies: map of development dependency name to version specifier

        """
        self._config.setdefault("tool", {})
        self._config["tool"].setdefault("poetry", {})
        self._config["tool"]["poetry"]["name"] = name
        self._config["tool"]["poetry"]["version"] = "1.0.0"
        self._config["tool"]["poetry"]["description"] = ""
        self._config["tool"]["poetry"]["authors"] = []
        self._config["tool"]["poetry"]["license"] = "Proprietary"
        self._config["tool"]["poetry"]["dependencies"] = dependencies
        self._config["tool"]["poetry"]["dev-dependencies"] = dev_dependencies
        self._config["build-system"] = {
            "requires": ["poetry-core>=1.0.0"],
            "build-backend": "poetry.core.masonry.api",
        }



"""Contains PythonRequirement class"""
from .base import BuildTarget


class PythonRequirement(BuildTarget):
    """Represents a Python requirement build target"""

    def __init__(self, package_name: str) -> None:
        """Initializer

        Args:
            package_name: package name

        """
        self.package_name = package_name
        self.key = f"{self.package_name}"



"""Contains functions for loading generator plugins"""
import importlib
import logging
import sys
from typing import Dict

from ..config import PROJECT_CONFIG  # noqa: I100
from .base import PackageGenerator  # noqa: I100

logger = logging.getLogger(__name__)


def load_project_generators() -> Dict[str, PackageGenerator]:
    """Load package generators defined in the project.

    Returns:
        dict of generator name (generator folder name) to package generator class

    """
    generators_path = PROJECT_CONFIG.config_dir_path.joinpath(".ns-poet/generators")
    generators = {}
    if generators_path.is_dir():
        sys.path.append(str(generators_path))
        for entry in generators_path.iterdir():
            if entry.is_dir():
                try:
                    mod = importlib.import_module(entry.name)
                    generators[entry.name] = mod.Generator  # type: ignore  # noqa
                except Exception as e:
                    logger.warning(
                        f"Failed to find a Generator class in {entry}: {str(e)}"
                    )

    return generators


# Create a singleton of package generator classes
PACKAGE_GENERATOR_MAP = load_project_generators()


def create_package_generator(
    package_type: str,
    nickname: str,
    title: str,
    description: str,
    top_level_dir: str = None,
    template: str = None,
) -> PackageGenerator:
    """Factory function for creating a package generator object.

    Args:
        package_type: Package type that corresponds to a key in the package generator map
        nickname: Nickname for the package. This will be used to folder names and
            other identifiers.
        title: Title of the package
        description: Short description of the package
        top_level_dir: The top level directory is where the package will be stored.
            e.g. apps. Providing this during instantiation and not as a class
            attribute should only be done for dynamic generator classes.
        template: Template directory path to provide if you don't want it be
            inferred. Providing this during instantiation and not as a class
            attribute should only be done for dynamic generator classes.

    Returns:
        new package generator object

    """
    PackageGeneratorClass: PackageGenerator = PACKAGE_GENERATOR_MAP[package_type]
    generator: PackageGenerator = PackageGeneratorClass(  # type: ignore  # noqa
        nickname, title, description, top_level_dir=top_level_dir, template=template
    )
    return generator



"""Contains the CLI"""

import logging
from pathlib import Path
import subprocess
from typing import Optional

import click

from ns_poet.processor import PackageProcessor
from ns_poet.project import PROJECT_CONFIG
from ns_poet.requirements import update_import_map

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    """Manage Poetry packages in a monorepo"""
    pass


@cli.group(name="import-map")
def import_map() -> None:
    """Commands for managing the import map"""
    pass


@import_map.command()
def update() -> None:
    """Update the import map from requirements.txt"""
    update_import_map()


@cli.group()
def package() -> None:
    """Commands for managing packages"""
    pass


@package.command()
@click.option(
    "-p",
    "--package-path",
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    help="Generate a package manifest for a single package path",
)
def generate(package_path: Optional[Path]) -> None:
    """Generate Poetry package manifests"""
    PROJECT_CONFIG.load_requirements()
    processor = PackageProcessor()
    processor.register_packages()
    processor.ensure_no_circular_imports()
    if package_path:
        processor.generate_package_manifest(package_path)
    else:
        processor.generate_package_manifests()


@package.command()
@click.option(
    "-p",
    "--package-path",
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    help="Install a single package at the given path",
)
def install(package_path: Optional[Path]) -> None:
    """Install Poetry packages"""
    if package_path:
        click.secho(f"Installing {package_path}...", fg="white", bold=True)
        subprocess.run(["poetry", "install"], cwd=str(package_path), check=True)
        click.secho(f"Installed {package_path}", fg="green", bold=True)
    else:
        for top_dir in PROJECT_CONFIG.top_dirs:
            for path in PROJECT_CONFIG.project_path.joinpath(top_dir).glob(
                "**/pyproject.toml"
            ):
                if any(
                    ignore_dir in str(path) for ignore_dir in PROJECT_CONFIG.ignore_dirs
                ):
                    click.secho(f"Ignoring {path.parent}...", fg="white")
                    continue

                click.secho(f"Installing {path.parent}...", fg="white", bold=True)
                subprocess.run(["poetry", "install"], cwd=str(path.parent), check=True)
                click.secho(f"Installed {path.parent}", fg="green", bold=True)



"""Utility functions"""

import ast
from pathlib import Path
import subprocess
from typing import Any, Set, Tuple


class ModuleImportVisitor(ast.NodeVisitor):
    """AST visitor for collecting imports from a module"""

    def __init__(self) -> None:
        """Initializer"""
        self.imports: Set[str] = set()

    def visit_ImportFrom(self, node: Any) -> None:
        """Collect an import when specified as `from foo import bar`"""
        if node.module is not None:
            self.imports.add(node.module.split(".")[0])

    def visit_Import(self, node: Any) -> None:
        """Collect an import when specified as `import foo`"""
        self.imports.add(node.names[0].name.split(".")[0])


def gather_dependencies_from_module(path: str) -> Tuple[str, Set[str]]:
    """Gather a set of other build targets from a single Python module

    Args:
        path: Python module path

    Returns:
        tuple of (path, set of package names that were imported in the module)

    """
    with open(path) as f:
        contents = f.read()
    tree = ast.parse(contents)
    visitor = ModuleImportVisitor()
    visitor.visit(tree)
    return path, visitor.imports


def get_git_top_level_path() -> Path:
    """Get the path of the git repository where this CLI is running"""
    proc = subprocess.run(
        ["git", "rev-parse", "--show-toplevel"], check=True, stdout=subprocess.PIPE
    )
    output = proc.stdout.decode("utf-8").strip()
    return Path(output)


def write_package_config_file(config_file_path: Path, contents: str) -> None:
    """Write package config file to disk

    Args:
        config_file_path: path to config file
        contents: TOML string to write

    """
    with config_file_path.open("w") as f:
        f.write(contents)



"""Contains a configuration parser and project config singleton"""
import json
from pathlib import Path
from typing import Any, Dict, FrozenSet, List, MutableMapping, Optional, Union

import pkg_resources
from pkg_resources import Requirement
import toml

from .util import get_git_top_level_path


class PoetProject:
    """Project object

    This object parses project configuration and manages the import map associated with
    the project's requirements.txt file.

    The schema for the configuration file in the git project root is::

        [tool.nspoet]
        ignore_dirs = []
        ignore_targets = []
        import_map_path = "3rdparty/python/import-map.json"
        requirements_path = "3rdparty/python/requirements.txt"
        top_dirs = ["."]
        default_python_version = "^3.6.5"

    """

    def __init__(self, project_path: Union[str, Path]) -> None:
        """Initializer

        Args:
            project_path: Path to the project

        """
        self.project_path = Path(project_path)
        self.config_file_path = self.project_path.joinpath("pyproject.toml")
        self._config: Optional[MutableMapping[str, Any]] = None
        self._requirements: Optional[Dict[str, Requirement]] = None
        self._import_map: Optional[Dict[str, str]] = None

    @classmethod
    def from_path(cls, project_path: Path) -> "PoetProject":
        """Create a project object and load configuration from a given project path

        Args:
            project_path: path to the project (git repo root)

        Returns:
            new PoetProject instance

        """
        p = cls(project_path)
        p.load_config()
        return p

    def load_config(self) -> MutableMapping[str, Any]:
        """Load configuration from disk

        Returns:
            configuration dict

        """
        if self.config_file_path.is_file():
            with self.config_file_path.open() as f:
                self._config = toml.load(f)
        else:
            self._config = {}

        return self._config

    @property
    def project_config(self) -> Dict[str, Any]:
        """Return the nspoet configuration subsection within the manifest"""
        return self._config.get("tool", {}).get("nspoet", {})

    @property
    def ignore_dirs(self) -> FrozenSet[str]:
        """Never look for or process files in this set of directories"""
        return frozenset(self.project_config.get("ignore_dirs", []))

    @property
    def ignore_targets(self) -> FrozenSet[str]:
        """Set of target package names to ignore when collecting build targets"""
        return frozenset(self.project_config.get("ignore_targets", []))

    @property
    def import_map_path(self) -> Path:
        """Path to the location of the import-map.json file within the project root

        This file contains a map from the import name found in *.py to the library's
        project name (i.e. requirements.txt name).

        """
        return self.project_path.joinpath(
            self.project_config.get(
                "import_map_path",
                "import-map.json",
            )
        )

    @property
    def requirements_path(self) -> Path:
        """Path to the requirements.txt within the project root"""
        return self.project_path.joinpath(
            self.project_config.get(
                "requirements_path",
                "requirements.txt",
            )
        )

    @property
    def top_dirs(self) -> List[str]:
        """Top-level directories to search for Python packages"""
        return self.project_config.get("top_dirs", ["."])

    @property
    def default_python_version(self) -> str:
        """Default python version to set in package manifests"""
        return self.project_config.get("default_python_version", "^3.6.5")

    def load_requirements(self) -> Dict[str, Requirement]:
        """Load requirement specifiers from disk

        Returns:
            map of requirement name/project name to requirement object

        """
        if not self._requirements:
            self._requirements = {}
            with self.requirements_path.open() as f:
                for r in pkg_resources.parse_requirements(f.read()):
                    self._requirements[r.unsafe_name] = r
                    self._requirements[r.project_name] = r

        return self._requirements

    def get_requirement(self, name: str) -> Requirement:
        """Get a requirement by name"""
        return self._requirements[name]

    def load_import_map(self) -> Dict[str, str]:
        """Load import map from disk

        Returns:
            map of import name to project name or empty dict if no file found

        """
        try:
            with self.import_map_path.open() as f:
                return json.load(f)
        except FileNotFoundError:
            return {}

    def get_import_map(self) -> Dict[str, str]:
        """Get the import map or load it if not already cached

        Returns:
            map of import name to project name

        """
        if not self._import_map:
            self._import_map = self.load_import_map()

        return self._import_map

    def update_import_map(self, import_map: Dict[str, str]) -> None:
        """Update import map file on disk

        Args:
            import_map: map of import name to project name

        """
        with self.import_map_path.open("w") as f:
            json.dump(import_map, f, sort_keys=True, indent=2)
            f.write("\n")


# Create a singleton for the project configuration
PROJECT_CONFIG = PoetProject.from_path(get_git_top_level_path())



"""Contains a configuration parser and project config singleton"""
from configparser import ConfigParser
import json
from pathlib import Path
from typing import Any, FrozenSet, List

from .util import get_git_top_level_path


class Config:
    """ns-poet configuration parser

    Configuration is based on a .ns-poet.cfg file.

    The schema for the file in the git project root is::

        [project]
        ignore_dirs = []
        ignore_targets = []
        python_package_name_prefix = ""
        top_dirs = ["."]

    The schema for the file in an individual python package is::

        [package]
        default_python_runtime = "python3.6"|"python3.7"|...

    Instantiating this class will load and initialize a config parser instance.
    """

    def __init__(self, config_dir_path: Path) -> None:
        """Initialize a config object

        Args:
            config_dir_path: path to a folder containing .ns-poet.cfg file

        """
        self.config_dir_path = config_dir_path
        self._poet_cfg_path = Path(config_dir_path).joinpath(".ns-poet.cfg")
        self._config = ConfigParser()
        self._config["project"] = {}
        self._config["package"] = {}
        if self._poet_cfg_path.is_file():
            self._config.read(self._poet_cfg_path)

    def set(self, *args: Any) -> None:
        """Proxy method to set a value on the underlying config parser instance"""
        self._config.set(*args)

    # Project options

    @property
    def ignore_dirs(self) -> FrozenSet[Any]:
        """Never look for or process files in these directories"""
        return frozenset(
            json.loads(self._config.get("project", "ignore_dirs", fallback="[]"))
        )

    @property
    def ignore_targets(self) -> FrozenSet[Any]:
        """Set of target package names to ignore when collecting build targets"""
        return frozenset(
            json.loads(self._config.get("project", "ignore_targets", fallback="[]"))
        )

    @property
    def python_package_name_prefix(self) -> str:
        """Prefix to use for names of packages generated by ns-poet"""
        return self._config.get("project", "python_package_name_prefix", fallback="")

    @property
    def top_dirs(self) -> List[str]:
        """Top-level directories to search for Python packages"""
        return json.loads(self._config.get("project", "top_dirs", fallback="['.']"))

    # Package options

    @property
    def default_python_runtime(self) -> str:
        """Default python runtime"""
        return self._config.get(
            "package", "default_python_runtime", fallback="python3.6"
        )


# Create a singleton for the project configuration
PROJECT_CONFIG = Config(get_git_top_level_path())



"""Contains the base class for a package generator"""
import inspect
import logging
import os
from pathlib import Path
import shutil
from typing import Dict, Optional

from cookiecutter.main import cookiecutter

from ..config import PROJECT_CONFIG

logger = logging.getLogger(__name__)


class PackageGenerator:
    """Base class for package generators.

    A package generator is responsible for creating a folder in the repo with standard
    boilerplate based on the type of package.

    """

    #: Package type must be defined in the child class
    PACKAGE_TYPE: str = None
    #: The top level directory is where the package will be stored. e.g. apps. It must
    #: be defined in the child class.
    TOP_LEVEL_DIR: str = None
    #: The folder containing the new package can have a prefix. This will depend on the
    #: package type.
    FOLDER_NAME_PREFIX = ""

    def __init__(
        self,
        nickname: str,
        title: str,
        description: str,
        top_level_dir: Optional[str] = None,
        template: Optional[str] = None,
    ) -> None:
        """Initializer

        Args:
            nickname: Nickname for the package. This will be used to folder names and
                other identifiers.
            title: Title of the package
            description: Short description of the package
            top_level_dir: The top level directory is where the package will be stored.
                e.g. apps. Providing this during instantiation and not as a class
                attribute should only be done for dynamic generator classes.
            template: Template directory path to provide if you don't want it be
                inferred. Providing this during instantiation and not as a class
                attribute should only be done for dynamic generator classes.

        """
        self.nickname = nickname
        self.title = title
        self.description = description
        self._top_level_dir = top_level_dir or self.TOP_LEVEL_DIR
        # The new package will be created in this output folder
        self._output_dir = PROJECT_CONFIG.config_dir_path.joinpath(self._top_level_dir)
        # The generator template should be defined in a folder in the same directory as
        # the generator module. Alternatively, it can be provided explicitly.
        self._template = template or os.path.dirname(inspect.getfile(self.__class__))

    @property
    def folder_name(self) -> str:
        """Basename of the folder for the new package"""
        return self.FOLDER_NAME_PREFIX + self.nickname

    @property
    def package_dir(self) -> str:
        """Absolute path to the new package directory"""
        return os.path.join(
            os.environ["TALOS_ROOT"], self._top_level_dir, self.folder_name
        )

    @property
    def package_name(self) -> str:
        """Name of the package. This is usually the name used for importing."""
        return self.folder_name

    @property
    def build_dir(self) -> str:
        """Directory containing the package's src files"""
        return os.path.join(self._top_level_dir, self.folder_name)

    @property
    def package_path(self) -> str:
        """Path to the package source code"""
        return os.path.join(self.package_dir, "src", self.package_name)

    @property
    def context(self) -> Dict:
        """Template context variables.

        Child classes can override this method and extend the dict.
        """
        return {
            "build_dir": self.build_dir,
            "description": self.description,
            "folder_name": self.folder_name,
            "package_name": self.package_name,
            "package_type": self.PACKAGE_TYPE,
            "title": self.title,
        }

    def generate(self) -> None:
        """Generate a new package in the repo"""
        try:
            cookiecutter(
                self._template,
                no_input=True,
                extra_context=self.context,
                replay=False,
                overwrite_if_exists=True,
                output_dir=self._output_dir,
            )
            self._post_generate()
        except Exception as e:
            logger.exception(f"Failed to generate the new package: {e}")
            shutil.rmtree(self.package_dir, ignore_errors=True)
            raise

    def _post_generate(self) -> None:
        """Optional post-generation step.

        This may be overridden by child classes.
        """
        pass

    def print_extra_help(self) -> None:
        """Print extra help info to the terminal

        This may be overridden by child classes.
        """
        pass


class PythonPackageGenerator(PackageGenerator):
    """Generator for a generic Python package.

    This class is meant to be extended.
    """

    PACKAGE_TYPE = "python"

    @property
    def folder_name(self) -> str:
        """Basename of the folder for the new package"""
        return super().folder_name.replace("-", "_")

    @property
    def package_name(self) -> str:
        """Name of the package. This is usually the name used for importing."""
        return f"{PROJECT_CONFIG.python_package_name_prefix}{self.folder_name}"

    @property
    def build_dir(self) -> str:
        """Directory containing the package's BUILD file"""
        return os.path.join(super().build_dir, "src")

    def _post_generate(self) -> None:
        """Cleanup the generated package"""
        self._clean_init_files()

    def _clean_init_files(self) -> None:
        """For some reason a couple __init__.py files get generated. Remove them."""
        try:
            root = Path(self.package_dir)
            files = list(root.rglob("*.pyc"))
            files.append(root.joinpath("__init__.py"))
            files.append(root.joinpath("src/__init__.py"))
            for f in files:
                f.unlink()
        except Exception:
            pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ns_poet', 'ns_poet.generators', 'ns_poet.package_targets']

package_data = \
{'': ['*']}

install_requires = \
['astor>=0.8.1,<0.9.0',
 'click>=8.0.3,<9.0.0',
 'networkx<2.6',
 'setuptools>=58.5.3,<59.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['nspoet = ns_poet.cli:cli']}

setup_kwargs = {
    'name': 'ns-poet',
    'version': '0.4.0',
    'description': 'Autogenerate Poetry package manifests in a monorepo',
    'long_description': '# ns-poet\n[![](https://img.shields.io/pypi/v/ns_poet.svg)](https://pypi.org/pypi/ns_poet/) [![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nManage Poetry packages in a monorepo\n\nFeatures:\n\n- Generate Poetry package manifests\n- Run a command in all packages\n\nTable of Contents:\n\n- [Installation](#installation)\n- [Guide](#guide)\n- [Development](#development)\n\n## Installation\n\nns-poet requires Python 3.6 or above.\n\n```bash\npip install ns-poet\n# or\npoetry add ns-poet\n```\n\n## Guide\n\n<!-- Subsections explaining how to use the package -->\n\n## Development\n\nTo develop ns-poet, install dependencies and enable the pre-commit hook:\n\n```bash\npip install pre-commit poetry\npoetry install\npre-commit install -t pre-commit -t pre-push\n```\n\nTo run tests:\n\n```bash\npoetry run pytest\n```\n',
    'author': 'Jonathan Drake',
    'author_email': 'jdrake@narrativescience.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NarrativeScience/ns-poet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.5,<4.0.0',
}


setup(**setup_kwargs)



"""Contains base BuildTarget class"""
from abc import ABC
from functools import total_ordering
from typing import Any


@total_ordering
class BuildTarget(ABC):
    """Represents a Python build target in Pants.

    This should be used as a mixin to provide common attributes and methods.
    """

    # This string will be used as an identifier for the build target
    key: str = None

    def __str__(self) -> str:
        """String representation"""
        return str(self.key)

    def __eq__(self, other: Any) -> bool:
        """Equality check"""
        if hasattr(other, "key"):
            return str(self.key) == str(other.key)
        else:
            return False

    def __hash__(self) -> int:
        """Object hash"""
        return hash((str(self.key),))

    def __repr__(self) -> str:
        """Object representation for debugging"""
        return str(self.key)

    def __lt__(self, other: Any) -> bool:
        """Less than comparator"""
        if hasattr(other, "key"):
            return str(self.key) < str(other.key)
        else:
            return False

    @property
    def dependency_target(self) -> str:
        """Returns the representation of this target in another target's dependencies"""
        return str(self.key)



"""Contains base PythonPackage class"""
import logging
import os
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from toml.decoder import InlineTableDict

from ns_poet.package import PoetPackage
from ns_poet.project import PROJECT_CONFIG
from .base import BuildTarget
from .requirement import PythonRequirement

logger = logging.getLogger(__name__)


class DynamicInlineTableDict(dict, InlineTableDict):
    """Subclass for inline tables in TOML"""


class PythonPackage(BuildTarget):
    """Represents a Python package (e.g. lib or test) build target in Pants"""

    def __init__(
        self,
        package_path: Union[str, Path],
    ) -> None:
        """Initializer

        Args:
            package_path: Path to the package directory, e.g. lib/python_core

        """
        self.package_path = Path(package_path)

        # This string will be used as an identifier for the build target
        self.key = str(package_path)

        # Set of dependencies to be gathered later
        self.dependencies: Set[BuildTarget] = set()
        self.dev_dependencies: Set[BuildTarget] = set()

        # Load the package config
        self.config = PoetPackage.from_path(self.package_path)

    @property
    def package_name(self) -> str:
        """Python package name

        Raises:
            :py:exc:`ValueError`: if no package name could be parsed from the directory
                structure

        Returns:
            package name string

        """
        for p in self.package_path.joinpath("src").iterdir():
            if p.is_dir() and not p.name.endswith(".egg-info"):
                return p.name

        raise ValueError("Could not determine package_name")

    def find_target_paths(self) -> List[Tuple["PythonPackage", str]]:
        """Find a list of target paths for parsing dependencies

        This default implementation collects all the .py files. Subclasses may override
        this method to collect different file types.

        Returns:
            list of tuples with items:
            * the current target (to indicate the owner)
            * the path to the file that will be parsed

        """
        target_paths = []
        for dirpath, dirnames, filenames in os.walk(self.package_path):
            if os.path.basename(dirpath) in PROJECT_CONFIG.ignore_dirs:
                # Empty the list of directories so os.walk does not recur
                dirnames.clear()
            else:
                for filename in filenames:
                    if filename.endswith(".py"):
                        target_paths.append((self, os.path.join(dirpath, filename)))

        return target_paths

    def add_dependency(
        self, targets: Dict[str, "PythonPackage"], module_path: str, package_name: str
    ) -> None:
        """Add a new dependency to the set

        A child class may choose to implement ``set_dependencies`` and skip this method.

        Args:
            targets: map of package name to instance of BuildTarget
            module_path: path of the module that was parsed for imports
            package_name: dependency package name, as parsed from an import statement

        """
        logger.debug(f"Processing {module_path}:{package_name}")
        if package_name in targets and package_name != self.package_name:
            if "tests" in Path(module_path).parts:
                self.dev_dependencies.add(targets[package_name])
                logger.debug(
                    f"Added dev dependency for existing target: {targets[package_name]}"
                )
            else:
                self.dependencies.add(targets[package_name])
                logger.debug(
                    f"Added dependency for existing target: {targets[package_name]}"
                )
            return

        import_map = PROJECT_CONFIG.get_import_map()
        if package_name in import_map:
            if "tests" in Path(module_path).parts:
                self.dev_dependencies.add(PythonRequirement(import_map[package_name]))
                logger.debug(
                    f"Added third-party dev dependency: {import_map[package_name]}"
                )
            else:
                self.dependencies.add(PythonRequirement(import_map[package_name]))
                logger.debug(
                    f"Added third-party dependency: {import_map[package_name]}"
                )

    def set_extra_dependencies(self, targets: Dict[str, "PythonPackage"]) -> None:
        """Set extra dependencies on the target

        This method allows targets to add dependencies based on their own custom logic.
        By default this is a no-op.

        A standard thing to do would be to call
        ``self.add_dependency(targets, package_name)`` one or more times.

        Args:
            targets: map of package name to instance of BuildTarget

        """
        pass

    def convert_dependencies(self, dependencies: Set[BuildTarget]) -> Dict[str, str]:
        """Convert dependencies into a map suitable for dumping to TOML format

        Args:
            dependencies: set of dependencies

        Raises:
            :py:exc:`NotImplementedError`: if dependency is not a requirement or Python
                package

        Returns:
            map of project name to version specifier

        """
        converted = {}
        for dependency in dependencies:
            if isinstance(dependency, PythonRequirement):
                requirement = PROJECT_CONFIG.get_requirement(dependency.key)
                version = str(requirement.specifier)  # type: ignore
                converted[requirement.unsafe_name] = version if version else "*"
            elif isinstance(dependency, PythonPackage):
                converted[dependency.package_name] = DynamicInlineTableDict(
                    path=os.path.relpath(dependency.package_path, self.package_path),
                    develop=True,
                )
            else:
                raise NotImplementedError(dependency)

        # Sort by key
        converted = {
            k: v for k, v in sorted(converted.items(), key=lambda item: item[0].lower())
        }

        return converted

    def generate_package_manifest(self) -> None:
        """Generate a Poetry package manifest for the build target"""
        dependencies = {"python": PROJECT_CONFIG.default_python_version}
        dependencies.update(self.convert_dependencies(self.dependencies))
        dev_dependencies = self.convert_dependencies(
            self.dev_dependencies - self.dependencies
        )
        self.config.update(self.package_name, dependencies, dev_dependencies)
